/* Talika panel applet: Switch between open windows using icons.
 * (C) 2009-2010 Sinew Software Systems
	 *
 * Authors:  Vinod Kumar Dahiya
 *           Hemant Kumar
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
	 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef TALIKA_LAUNCHER_H__
#define TALIKA_LAUNCHER_H__

#include <gtk/gtk.h>
#include <gio/gio.h>

#include "task.h"
#include "tasklist.h"

typedef struct _TalikaIconCache
{
	GHashTable *hash;
}
TalikaIconCache;

TalikaIconCache* 
talika_icon_cache_new(void);

void
talika_icon_cache_free(TalikaIconCache *cache);

void
talika_icon_cache_add (TalikaIconCache *cache,
                       const gchar *desktop_id,
                       GdkPixbuf *icon);

GdkPixbuf*
talika_icon_cache_get_icon (TalikaIconCache *cache,
                                      const gchar *desktop_id);


typedef struct
{
	gchar *class_name;
	gchar *desktop_id;
}
TalikaClass;

typedef struct _TalikaClassCache
{
	GHashTable *hash;
	GHashTable *id_hash; //desktop id
}
TalikaClassCache;

TalikaClassCache* 
talika_class_cache_new(void);

void
talika_class_cache_free(TalikaClassCache *cache);

void
talika_class_cache_update_window (TalikaClassCache *cache,
                                          WnckWindow *window);

TalikaClass*
talika_class_cache_get_class (TalikaClassCache *cache,
                                      WnckWindow *window);

void
talika_class_cache_free_window (TalikaClassCache *cache,
                                        WnckWindow *window);

GdkPixbuf*
talika_get_icon_from_desktop_file(TalikaTaskList *tasklist, const gchar *desktop_id, guint size);

gchar *talika_get_spec_from_class_pid (const gchar *name, gint pid);

struct _TalikaTask *talika_launcher_new(TalikaTaskList *tasklist,
                                    const gchar *spec);

void talika_launcher_free(struct _TalikaTask *launcher);

void talika_launcher_launch (GtkWidget *menu_item, struct _TalikaTask *launcher);

void talika_launcher_unpin (GtkWidget *menu_item, struct _TalikaTask *launcher);

void do_talika_launcher_update_visible_state (gpointer data, 
                                     gpointer user_data);

GdkPixbuf* 
talika_launcher_get_icon (struct _TalikaTask *launcher);

void 
talika_launcher_launch_uris (GtkWidget *menu_item, 
                             struct _TalikaTask *launcher,
                             GList *uris);

void 
talika_launcher_change_icon (GtkWidget *menu_item, struct _TalikaTask *launcher);

#endif //TALIKA_LAUNCHER_H__
