
/* Talika panel applet: Switch between open windows using icons.
 * (C) 2009-2010 Sinew Software Systems
	 *
 * Authors:  Vinod Kumar Dahiya
 *           Hemant Kumar
 *
 * NOTE : A lot of code ( about 90% ) has been taken from libwnck tasklist.c
	 * we just renamed & modified it to suit our purpose.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as pub lished by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
	 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <libwnck/libwnck.h>
#include <math.h>
#include <panel-applet-gconf.h>
#include <gconf/gconf-client.h>

#include "tasklist.h"
#include "task.h"
#include "xutils.h"
#include "tooltips.h"
#include "launcher.h"



static void 
talika_tasklist_active_window_changed (WnckScreen * screen,
                                       WnckWindow * previous_window,
                                       TalikaTaskList * tasklist);

static GtkWidget* 
get_talika_tasklist_box (PanelAppletOrient orient);

static gboolean 
talika_tasklist_expose (GtkWidget * widget,
                        GdkEventExpose * event, gpointer data);



/*####################### Active window stuff ##############################*/

static gboolean 
talika_tasklist_change_active_timeout (gpointer data)
{
	TalikaTaskList *tasklist = data;

	tasklist->activate_timeout_id = 0;

	talika_tasklist_active_window_changed (tasklist->screen, NULL,
	                                       tasklist);

	return FALSE;
}

void 
talika_tasklist_activate_task_window (TalikaTask * task, 
                                      guint32 timestamp, gboolean force_maximize)
{
	TalikaTaskList *tasklist;
	WnckWindowState state;
	WnckWorkspace *active_ws;
	WnckWorkspace *window_ws;

	tasklist = task->tasklist;

	if (task->window == NULL)
		return;

	state = wnck_window_get_state (task->window);

	active_ws = wnck_screen_get_active_workspace (tasklist->screen);
	window_ws = wnck_window_get_workspace (task->window);

	if (state & WNCK_WINDOW_STATE_MINIMIZED||force_maximize)
	{
		if (window_ws &&
		    active_ws != window_ws &&
		    !tasklist->switch_workspace_on_unminimize)
			wnck_workspace_activate (window_ws, timestamp);

		wnck_window_activate_transient (task->window, timestamp);
	}
	else
	{
		if ((task->was_active ||
		     wnck_window_transient_is_most_recently_activated
		     (task->window)) && (!window_ws
		                         || active_ws == window_ws))
		{
			//update cache first
			talika_preview_image_cache_update_window (task->tasklist->preview_cache,
			                                          task->window,
			                                          task->tasklist->preview_size,
			                                          task->tasklist->box);
			
			task->was_active = FALSE;
			wnck_window_minimize (task->window);
			return;
		}
		else
		{
			// FIXME: THIS IS SICK AND WRONG AND BUGGY.  See the end of
			// http://mail.gnome.org/archives/wm-spec-list/2005-July/msg00032.html
			// There should only be *one* activate call.
			//
			if (window_ws)
				wnck_workspace_activate (window_ws,
				                         timestamp);

			wnck_window_activate_transient (task->window,
			                                timestamp);
		}
	}


	if (tasklist->activate_timeout_id)
		g_source_remove (tasklist->activate_timeout_id);

	tasklist->activate_timeout_id =
		g_timeout_add (500, &talika_tasklist_change_active_timeout,
		               tasklist);

	talika_tasklist_change_active_task (tasklist, task);
}

void 
talika_tasklist_change_active_task (TalikaTaskList * tasklist,
                                    TalikaTask * active_task)
{
	if (active_task && active_task == tasklist->active_task)
		return;

	if (tasklist->active_task)
	{
		tasklist->active_task->really_toggling = TRUE;
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
		                              (tasklist->active_task->button),
		                              FALSE);
		tasklist->active_task->really_toggling = FALSE;
	}

	tasklist->active_task = active_task;

	if (tasklist->active_task)
	{
		tasklist->active_task->really_toggling = TRUE;
		if (tasklist->toggle_current_task)
			gtk_toggle_button_set_active
			(GTK_TOGGLE_BUTTON
			 (tasklist->active_task->button), TRUE);
		tasklist->active_task->really_toggling = FALSE;
	}

	//update group button state
	if (active_task)
	{
		active_task =
			g_hash_table_lookup (tasklist->class_group_hash,
			                     active_task->class_group_name);
	}

	if (active_task &&
	    active_task == tasklist->active_class_group)
		return;

	if (tasklist->active_class_group)
	{
		tasklist->active_class_group->really_toggling = TRUE;
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
		                              (tasklist->
		                               active_class_group->button),
		                              FALSE);
		tasklist->active_class_group->really_toggling = FALSE;
	}

	tasklist->active_class_group = active_task;

	if (tasklist->active_class_group)
	{
		tasklist->active_class_group->really_toggling = TRUE;
		if (tasklist->toggle_current_task)
			gtk_toggle_button_set_active
			(GTK_TOGGLE_BUTTON
			 (tasklist->
			  active_class_group->button), TRUE);
		tasklist->active_class_group->really_toggling = FALSE;
	}
	
}

static void 
talika_tasklist_active_window_changed (WnckScreen * screen,
                                       WnckWindow *
                                       previous_window,
                                       TalikaTaskList * tasklist)
{
	WnckWindow *active_window;
	WnckWindow *initial_window;
	TalikaTask *active_task = NULL;

	// FIXME: check for group modal window 
	initial_window = active_window =
		wnck_screen_get_active_window (screen);
	active_task = g_hash_table_lookup (tasklist->win_hash, active_window);
	while (active_window && !active_task)
	{
		active_window = wnck_window_get_transient (active_window);
		active_task =
			g_hash_table_lookup (tasklist->win_hash,
			                     active_window);
		// Check for transient cycles 
		if (active_window == initial_window)
			break;
	}

	talika_tasklist_change_active_task (tasklist, active_task);
}


/*########################## Update Tasks List ###############################*/

static void 
talika_tasklist_free_skipped_windows (TalikaTaskList * tasklist)
{
	GList *l;

	l = tasklist->skipped_windows;

	while (l != NULL)
	{
		TalikaSkippedWindow *skipped =
			(TalikaSkippedWindow *) l->data;
		g_signal_handler_disconnect (skipped->window, skipped->tag);
		g_free (skipped);
		l = l->next;
	}

	g_list_free (tasklist->skipped_windows);
}

static void 
talika_tasklist_free_tasks (TalikaTaskList * tasklist)
{
	GList *l;

	tasklist->active_task = NULL;
	tasklist->active_class_group = NULL;

	if (tasklist->windows)
	{
		l = tasklist->windows;
		while (l != NULL)
		{
			TalikaTask *task = (TalikaTask *) l->data;
			l = l->next;
			talika_task_free (task);
		}

		g_list_free (tasklist->windows);
		tasklist->windows = NULL;

		g_hash_table_remove_all (tasklist->win_hash);
	}

	g_list_free (tasklist->windows_without_class_group);
	tasklist->windows_without_class_group = NULL;

	if (tasklist->class_groups)
	{
		l = tasklist->class_groups;
		while (l != NULL)
		{
			TalikaTask *task = (TalikaTask *) l->data;
			l = l->next;
			talika_task_free (task);
		}

		g_list_free (tasklist->class_groups);
		tasklist->class_groups = NULL;

		g_hash_table_remove_all (tasklist->class_group_hash);
	}

	if (tasklist->skipped_windows)
	{
		talika_tasklist_free_skipped_windows (tasklist);
		tasklist->skipped_windows = NULL;
	}

	g_list_free (tasklist->visible_tasks);
	tasklist->visible_tasks = NULL;
}

static void 
talika_tasklist_free_launchers (TalikaTaskList * tasklist)
{
	GList *l;

	if (tasklist->launchers)
	{
		l = tasklist->launchers;
		while (l != NULL)
		{
			TalikaTask *launcher = (TalikaTask *) l->data;
			l = l->next;
			talika_launcher_free (launcher);
		}

		g_list_free (tasklist->launchers);
		tasklist->launchers = NULL;
	}
}

static gint 
talika_tasklist_windows_compare (gconstpointer a, 
                                 gconstpointer b)
{
	int posa;
	int posb;

	posa = wnck_window_get_sort_order (WNCK_WINDOW (a));
	posb = wnck_window_get_sort_order (WNCK_WINDOW (b));

	if (posa < posb)
		return -1;
	else if (posa > posb)
		return 1;

	return 0;
}

static gint 
talika_tasklist_tasks_compare (gconstpointer a, 
                                 gconstpointer b)
{
	int posa;
	int posb;

	posa = talika_task_get_sort_order ((TalikaTask*)a);
	posb = talika_task_get_sort_order ((TalikaTask*)b);

	if (posa < posb)
		return -1;
	else if (posa > posb)
		return 1;

	return 0;
}

static void 
talika_tasklist_unparent_buttons (GtkWidget *child, gpointer parent)
{
	gtk_container_remove (GTK_CONTAINER(parent),child);
}

static void 
talika_tasklist_update_visible_tasks (TalikaTaskList * tasklist)
{	
	//unparent alredy added launchers, and reinitlise
	gtk_container_foreach (GTK_CONTAINER(tasklist->box),
	                       talika_tasklist_unparent_buttons,
	                       tasklist->box);
	GList *ls;
	for (ls = tasklist->launchers; ls; ls = ls->next)
	{
		TalikaTask *launcher = ls->data;

		launcher->exclude = FALSE;
		launcher->attached_task = NULL;

		gtk_widget_set_state (launcher->button, GTK_STATE_NORMAL);
	}

	//update visible tasks
	if (tasklist->use_grouping)
	{
		GList *l;
		for (l = tasklist->class_groups; l; l = l->next)
		{
			TalikaTask *class_group_task = l->data;
			if (g_list_length (class_group_task->windows) > 1)
			{
				tasklist->visible_tasks =
					g_list_prepend
					(tasklist->visible_tasks,
					 class_group_task);
			}
			else
			{
				TalikaTask *task =
					class_group_task->windows->data;
				tasklist->visible_tasks =
					g_list_prepend
					(tasklist->visible_tasks, task);
			}
		}

		for (l = tasklist->windows_without_class_group; l;
		     l = l->next)
		{
			TalikaTask *task = l->data;
			tasklist->visible_tasks =
				g_list_prepend
				(tasklist->visible_tasks, task);
		}
	}
	else
	{
		tasklist->visible_tasks = g_list_copy (tasklist->windows);
	}

	/************* add launchers to box first ***************/

	//connect launchers to tasks
	GList *l;
	for (l = tasklist->visible_tasks; l; l = l->next)
	{
		TalikaTask *task = l->data;

		GList *ls;
		for (ls = tasklist->launchers; ls; ls = ls->next)
		{
			TalikaTask *launcher = ls->data;

			if (strcmp(launcher->class_name, task->class_group_name) == 0)
			{
				task->launcher = launcher;

				if (!launcher->exclude)
				{
					launcher->exclude = TRUE;
					talika_task_set_sort_order 
						(task, talika_task_get_sort_order (launcher));
					launcher->attached_task = task;
				}
				
				break;
			}
			else if (strcmp(launcher->class_name, "nil") == 0)
			{
				//launcher do not have valid class name, try to find
				WnckWindow *window=NULL;
				switch (task->type)
				{
					case TALIKA_TASK_WINDOW:
						window = task->window;
						break;

					case TALIKA_TASK_CLASS_GROUP:
						window = 
							((TalikaTask *)(task->windows->data))->window;
						break;

					default:
						break;
				}

				TalikaClass *class_obj = 
					talika_class_cache_get_class 
					(task->tasklist->class_cache, window);

				//assign class if desktop_id matches
				if (class_obj->desktop_id && 
				    strcmp(launcher->desktop_id, class_obj->desktop_id) == 0)
				{
					g_free(launcher->class_name);
					launcher->class_name = g_strdup (class_obj->class_name);
					talika_tasklist_save_launchers (launcher->tasklist);

					task->launcher = launcher;

					if (!launcher->exclude)
					{
						launcher->exclude = TRUE;
						talika_task_set_sort_order 
							(task, talika_task_get_sort_order (launcher));
						launcher->attached_task = task;
					}

					break;
				}
			}
		}
	}

	GList *lun;
	for (lun = tasklist->launchers; lun; lun = lun->next)
	{
		TalikaTask *launcher = lun->data;

		if (!launcher->exclude)
		{
			tasklist->visible_tasks = g_list_prepend(tasklist->visible_tasks,
			                                        launcher);
		}
	}

	tasklist->visible_tasks = g_list_sort (tasklist->visible_tasks,
	                                       talika_tasklist_tasks_compare); 
	
	/****************** add task to box now*********************/
	GList *l2;
	for (l2 = tasklist->visible_tasks; l2; l2 = l2->next)
	{
		TalikaTask *task = l2->data;
		gtk_box_pack_start (GTK_BOX (tasklist->box),
		                    task->button, FALSE, FALSE, 1);
	}

	//adopt the custom icons for launchers now
	g_list_foreach (tasklist->visible_tasks,
	                do_talika_task_update_visible_state, NULL);
}

static void 
talika_tasklist_update_icon_geometries (TalikaTaskList *tasklist,
                                        GList * visible_tasks)
{
	gint x, y, width, height;
	GList *l1;

	for (l1 = visible_tasks; l1; l1 = l1->next)
	{
		TalikaTask *task = l1->data;

		if (!GTK_WIDGET_REALIZED (task->button))
			continue;

		gdk_window_get_origin (GTK_BUTTON
		                       (task->button)->event_window, &x, &y);
		width = task->button->allocation.width;
		height = task->button->allocation.height;

		if (task->type == TALIKA_TASK_WINDOW)
			wnck_window_set_icon_geometry (task->window,
			                               x, y, width, height);
		else
		{
			GList *l;

			for (l = task->windows; l; l = l->next)
			{
				TalikaTask *task1 = l->data;
				wnck_window_set_icon_geometry (task1->window,
				                               x, y, width,
				                               height);
			}
		}
	}
}

/*
 * This function determines if a window should be included in the tasklist.
 */
static gboolean 
tasklist_include_window_impl (TalikaTaskList * tasklist,
                              WnckWindow * win,
                              gboolean check_for_skipped_list)
{
	WnckWorkspace *active_workspace;
	int x, y, w, h;

	if (!check_for_skipped_list &&
	    wnck_window_get_state (win) & WNCK_WINDOW_STATE_SKIP_TASKLIST)
		return FALSE;

	if (tasklist->monitor_num != -1)
	{
		wnck_window_get_geometry (win, &x, &y, &w, &h);
		// Don't include the window if its center point 
		// is not on the same monitor 
		if (gdk_screen_get_monitor_at_point
		    (talika_screen_get_gdk_screen (tasklist->screen),
		     x + w / 2, y + h / 2) != tasklist->monitor_num)
			return FALSE;
	}

	// Remainder of checks aren't relevant for checking if the window should
	// be in the skipped list.

	if (check_for_skipped_list)
		return TRUE;

	if (tasklist->include_all_workspaces)
		return TRUE;

	if (wnck_window_is_pinned (win))
		return TRUE;

	active_workspace =
		wnck_screen_get_active_workspace (tasklist->screen);
	if (active_workspace == NULL)
		return TRUE;

	if (wnck_window_or_transient_needs_attention (win))
		return TRUE;

	if (active_workspace != wnck_window_get_workspace (win))
		return FALSE;

	if (!wnck_workspace_is_virtual (active_workspace))
		return TRUE;

	return wnck_window_is_in_viewport (win, active_workspace);
}

static gboolean 
tasklist_include_in_skipped_list (TalikaTaskList * tasklist,
                                  WnckWindow * win)
{
	return tasklist_include_window_impl (tasklist, win, TRUE);
}

static gboolean 
talika_tasklist_include_window (TalikaTaskList * tasklist, 
                                WnckWindow * win)
{
	return tasklist_include_window_impl (tasklist, win, FALSE);
}

/* This is main function build list of tasks */
void 
talika_tasklist_update_lists (TalikaTaskList * tasklist)
{
	if (tasklist->enter_timeout_id)
		g_source_remove (tasklist->enter_timeout_id);
	tasklist->enter_timeout_id =0;
	
	
	g_print ("updating list ...\n");

	GList *windows;
	WnckWindow *win;
	WnckClassGroup *class_group;
	GList *l;
	TalikaTask *win_task;
	TalikaTask *class_group_task;

	//free all tasks prerior to rebuild
	talika_tasklist_free_tasks (tasklist);

	//update monitor number and geomatry
	if (GTK_WIDGET (tasklist->box)->window != NULL)
	{
		if(tasklist->include_all_monitors)
		{
			tasklist->monitor_num = -1;
		}
		else
		{
			int monitor_num;

			monitor_num =
				gdk_screen_get_monitor_at_window
				(talika_screen_get_gdk_screen
				 (tasklist->screen),
				 GTK_WIDGET (tasklist->box)->window);

			if (monitor_num != tasklist->monitor_num)
			{
				tasklist->monitor_num = monitor_num;
				gdk_screen_get_monitor_geometry
					(talika_screen_get_gdk_screen
					 (tasklist->screen),
					 tasklist->monitor_num,
					 &tasklist->monitor_geometry);
			}
		}
	}

	//get sorted copy of window lists on screen
	l = g_list_copy (wnck_screen_get_windows (tasklist->screen));
	windows = l = g_list_sort (l, talika_tasklist_windows_compare);

	//populate windows, class and skipped list
	while (l != NULL)
	{
		win = WNCK_WINDOW (l->data);

		if (talika_tasklist_include_window (tasklist, win))
		{
			TalikaClass *class_obj = 
				talika_class_cache_get_class (tasklist->class_cache,win);

			if(!class_obj)
			{
				talika_class_cache_update_window (tasklist->class_cache,
	                                          win);
				class_obj = 
					talika_class_cache_get_class (tasklist->class_cache,win);
			}
			
			win_task =
				talika_task_new_from_window (tasklist, win, 
				                             class_obj->class_name);
			tasklist->windows =
				g_list_append (tasklist->windows, win_task);
			g_hash_table_insert (tasklist->win_hash, win,
			                     win_task);

			gtk_widget_show (win_task->button);

			// Class group 

			class_group = wnck_window_get_class_group (win);
			
			// don't group windows if they do not belong to any class 
			if (strcmp
			    (class_obj->class_name,
			     "") != 0)
			{
				class_group_task =
					g_hash_table_lookup
					(tasklist->class_group_hash,
					 class_obj->class_name);

				if (class_group_task == NULL)
				{
					class_group_task =
						talika_task_new_from_class_group
						(tasklist, class_group, class_obj->class_name);
					gtk_widget_show
						(class_group_task->button);

					tasklist->class_groups =
						g_list_prepend
						(tasklist->class_groups,
						 class_group_task);
					
					g_hash_table_insert
						(tasklist->class_group_hash,
						 class_group_task->class_group_name,
						 class_group_task);
				}

				//class grouping order
				if(tasklist->group_reverse_order)
				{
					class_group_task->windows =
						g_list_append
						(class_group_task->windows, win_task);
				}
				else
				{
					class_group_task->windows =
						g_list_prepend
						(class_group_task->windows, win_task);
				}
			}
			else
			{
				//g_object_ref (win_task);
				tasklist->windows_without_class_group =
					g_list_prepend
					(tasklist->
					 windows_without_class_group,
					 win_task);
			}
		}
		else if (tasklist_include_in_skipped_list (tasklist, win))
		{
			TalikaSkippedWindow *skipped =
				g_new0 (TalikaSkippedWindow, 1);
			skipped->window = win;
			skipped->tag =
				g_signal_connect (G_OBJECT (win),
				                  "state_changed",
				                  G_CALLBACK
				                  (talika_task_state_changed),
				                  tasklist);
			tasklist->skipped_windows =
				g_list_prepend (tasklist->skipped_windows,
				                (gpointer) skipped);
		}

		l = l->next;
	}

	g_list_free (l);

	/* Sort the class group list */
	/*l = tasklist->class_groups;
	while (l)
	{
		class_group_task = (TalikaTask *) l->data;

		//original libwnck sorts class windows in alphabatical order
		//we are not using it 
		//class_group_task->windows =
		// g_list_sort (class_group_task->windows,
		//                 wnck_task_compare); 

		// so the number of windows in the task gets reset on the
		// task label or tooltip
		//talika_task_update_visible_state (class_group_task);

		l = l->next;
	}*/

	/* update visible task list that should be actually shown*/ 
	talika_tasklist_update_visible_tasks (tasklist);

	/* since we cleared active_window we need to reset it */
	talika_tasklist_active_window_changed (tasklist->screen, NULL,
	                                       tasklist);

	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

static gboolean 
do_talika_tasklist_update_lists (gpointer data)
{
	TalikaTaskList *tasklist = (TalikaTaskList *) data;

	tasklist->idle_callback_tag = 0;

	talika_tasklist_update_lists (tasklist);

	return FALSE;
}


/*################ Various Screen, workspace, window notifications ###########*/

static void 
talika_tasklist_window_changed_workspace (WnckWindow * window,
                                          TalikaTaskList * tasklist)
{
	WnckWorkspace *active_ws;
	WnckWorkspace *window_ws;
	gboolean need_update;
	GList *l;

	active_ws = wnck_screen_get_active_workspace (tasklist->screen);
	window_ws = wnck_window_get_workspace (window);

	if (!window_ws)
		return;

	need_update = FALSE;

	if (active_ws == window_ws)
		need_update = TRUE;

	l = tasklist->windows;
	while (!need_update && l != NULL)
	{
		TalikaTask *task = l->data;

		if (task->type == TALIKA_TASK_WINDOW &&
		    task->window == window)
			need_update = TRUE;

		l = l->next;
	}

	if (need_update)
	{
		talika_tasklist_update_lists (tasklist);
		//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
	}
}

static void 
talika_tasklist_window_changed_geometry (WnckWindow * window,
                                         TalikaTaskList * tasklist)
{
	TalikaTask *win_task;
	gboolean show;
	gboolean monitor_changed;
	int x, y, w, h;

	if (tasklist->idle_callback_tag != 0)
		return;

	/*
	 * If the (parent of the) tasklist itself skips
		 * the tasklist, we need an extra check whether
	 * the tasklist itself possibly changed monitor.
	 */
	monitor_changed = FALSE;
	if (tasklist->monitor_num != -1 &&
	    (wnck_window_get_state (window) &
	     WNCK_WINDOW_STATE_SKIP_TASKLIST)
	    && GTK_WIDGET (tasklist->box)->window != NULL)
	{
		/* Do the extra check only if there is a suspect of a monitor change
		 (= this window is off monitor) */
		wnck_window_get_geometry (window, &x, &y, &w, &h);
		if (!POINT_IN_RECT
		    (x + w / 2, y + h / 2, tasklist->monitor_geometry))
		{
			monitor_changed =
				(gdk_screen_get_monitor_at_window
				 (talika_screen_get_gdk_screen
				  (tasklist->screen),
				  GTK_WIDGET (tasklist->box)->window) !=
				 tasklist->monitor_num);
		}
	}

	/*
	 * We want to re-generate the task list if
	 * the window is shown but shouldn't be or
	 * the window isn't shown but should be or
	 * the tasklist itself changed monitor.
	 */
	win_task = g_hash_table_lookup (tasklist->win_hash, window);
	show = talika_tasklist_include_window (tasklist, window);
	if (((win_task == NULL && !show)
	     || (win_task != NULL && show)) && !monitor_changed)
		return;

	/* Don't keep any stale references */
	gtk_widget_queue_draw (GTK_WIDGET (tasklist->box));

	tasklist->idle_callback_tag =
		g_idle_add (do_talika_tasklist_update_lists, tasklist);
}

static void 
talika_tasklist_connect_window (TalikaTaskList * tasklist,
                                WnckWindow * window)
{
	g_signal_connect (window, "workspace_changed",
	                  G_CALLBACK
	                  (talika_tasklist_window_changed_workspace),
	                  tasklist);
	g_signal_connect (window, "geometry_changed",
	                  G_CALLBACK
	                  (talika_tasklist_window_changed_geometry),
	                  tasklist);
}

static void 
talika_tasklist_disconnect_window (TalikaTaskList * tasklist,
                                   WnckWindow * window)
{
	g_signal_handlers_disconnect_by_func (window,
	                                      talika_tasklist_window_changed_workspace,
	                                      tasklist);
	g_signal_handlers_disconnect_by_func (window,
	                                      talika_tasklist_window_changed_geometry,
	                                      tasklist);
}

static void 
talika_tasklist_window_added (WnckScreen * screen,
                              WnckWindow * win, 
                              TalikaTaskList * tasklist)
{
	g_print ("window added: ");
	//g_print ("%s : sort : %d",wnck_window_get_name (win),
	//         wnck_window_get_sort_order (win));
	g_print ("\n");
	
	talika_tasklist_connect_window (tasklist, win);

	//update class cache
	talika_class_cache_update_window (tasklist->class_cache,
	                                          win);
	
	//update preview cache
	talika_preview_image_cache_update_window (tasklist->preview_cache,
	                                          win,
	                                          tasklist->preview_size,
	                                          tasklist->box);
	
	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

static void 
talika_tasklist_window_removed (WnckScreen * screen,
                                WnckWindow * win, 
                                TalikaTaskList * tasklist)
{
	g_print ("window removed\n");
	
	talika_tasklist_disconnect_window (tasklist, win);	

	//remove window from preview cache
	talika_preview_image_cache_free_window (tasklist->preview_cache,
	                                          win);

	//remove window from class cache
	talika_class_cache_free_window (tasklist->class_cache,
	                                          win);
	
	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

static void 
talika_tasklist_active_workspace_changed (WnckScreen * screen,
                                          WnckWorkspace *
                                          previous_workspace,
                                          TalikaTaskList * tasklist)
{
	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

static void 
talika_tasklist_viewports_changed (WnckScreen * screen,
                                   TalikaTaskList * tasklist)
{
	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

static void 
talika_tasklist_connect_screen (TalikaTaskList * tasklist)
{
	GList *windows;
	guint *c;
	int i;
	WnckScreen *screen;

	g_return_if_fail (tasklist->screen != NULL);

	screen = tasklist->screen;

	i = 0;
	c = tasklist->screen_connections;

	c[i++] = g_signal_connect (G_OBJECT (screen),
	                           "active_window_changed",
	                           G_CALLBACK
	                           (talika_tasklist_active_window_changed),
	                           tasklist);
	c[i++] = g_signal_connect (G_OBJECT (screen),
	                           "active_workspace_changed",
	                           G_CALLBACK
	                           (talika_tasklist_active_workspace_changed),
	                           tasklist);
	c[i++] = g_signal_connect (G_OBJECT (screen), "window_opened",
	                           G_CALLBACK
	                           (talika_tasklist_window_added), tasklist);
	c[i++] = g_signal_connect (G_OBJECT (screen), "window_closed",
	                           G_CALLBACK
	                           (talika_tasklist_window_removed),
	                           tasklist);
	c[i++] = g_signal_connect (G_OBJECT (screen),
	                           "viewports-changed",
	                           G_CALLBACK
	                           (talika_tasklist_viewports_changed),
	                           tasklist);
	g_assert (i == N_SCREEN_CONNECTIONS);

	windows = wnck_screen_get_windows (screen);
	while (windows != NULL)
	{
		talika_tasklist_connect_window (tasklist, windows->data);
		windows = windows->next;
	}
}

static void 
talika_tasklist_disconnect_screen (TalikaTaskList * tasklist)
{
	GList *windows;
	int i;
	guint *c;
	WnckScreen *screen;

	g_return_if_fail (tasklist->screen != NULL);

	screen = tasklist->screen;

	windows = wnck_screen_get_windows (screen);
	while (windows != NULL)
	{
		talika_tasklist_disconnect_window (tasklist, windows->data);
		windows = windows->next;
	}

	i = 0;
	c = tasklist->screen_connections;
	while (i < N_SCREEN_CONNECTIONS)
	{
		if (c[i] != 0)
			g_signal_handler_disconnect (G_OBJECT (screen), c[i]);

		c[i] = 0;

		++i;
	}

	g_assert (i == N_SCREEN_CONNECTIONS);
}


/*###################### Prefrences changed callbacks ########################*/

void 
talika_tasklist_update_button_spacing (TalikaTaskList * tasklist)
{
	gtk_box_set_spacing(GTK_BOX(tasklist->box),tasklist->button_spacing);
	//talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

void 
talika_tasklist_update_visible_states (TalikaTaskList * tasklist)
{
	g_list_foreach (tasklist->launchers,
		            do_talika_task_update_visible_state, NULL);
	
	g_list_foreach (tasklist->visible_tasks,
	                do_talika_task_update_visible_state, NULL);
}

void 
talika_tasklist_update_icon_size (TalikaTaskList * tasklist)
{
	//reset icon cache
	talika_icon_cache_free(tasklist->icon_cache);
	tasklist->icon_cache = talika_icon_cache_new ();

	g_list_foreach (tasklist->launchers,
	                do_talika_task_update_visible_state, NULL);

	g_list_foreach (tasklist->visible_tasks,
	                do_talika_task_update_visible_state, NULL);
}

void 
talika_tasklist_update_mono_icons (TalikaTaskList * tasklist)
{
	g_list_foreach (tasklist->launchers,
		            do_talika_launcher_update_visible_state, NULL);
	
	g_list_foreach (tasklist->visible_tasks,
	                do_talika_task_update_visible_state, NULL);
}

void 
talika_tasklist_update_enable_previews (TalikaTaskList * tasklist)
{
	//g_list_foreach (tasklist->visible_tasks,
	//                do_talika_task_update_visible_state, NULL);
	if(!tasklist->enable_previews && tasklist->preview_cache)
	{
		talika_preview_image_cache_free (tasklist->preview_cache);
		tasklist->preview_cache = NULL;
	}
	else if(tasklist->enable_previews && !tasklist->preview_cache)
	{
		tasklist->preview_cache = talika_preview_image_cache_new ();
	}
}

void 
talika_tasklist_update_show_all_workspaces (TalikaTaskList * tasklist)
{
	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

void 
talika_tasklist_update_group_windows (TalikaTaskList * tasklist)
{
	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

void 
talika_tasklist_update_button_relief (TalikaTaskList * tasklist)
{
	g_list_foreach (tasklist->visible_tasks,
	                do_talika_task_update_visible_state, NULL);
}

void 
talika_tasklist_update_toggle_current_task (TalikaTaskList * tasklist)
{
	if (tasklist->active_task)
	{
		tasklist->active_task->really_toggling = TRUE;
		if (tasklist->toggle_current_task)
			gtk_toggle_button_set_active
			(GTK_TOGGLE_BUTTON
			 (tasklist->active_task->button), TRUE);
		else
			gtk_toggle_button_set_active
			(GTK_TOGGLE_BUTTON
			 (tasklist->active_task->button), FALSE);
		tasklist->active_task->really_toggling = FALSE;

		//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
	}
}

void 
talika_tasklist_update_dimm_minimised (TalikaTaskList * tasklist)
{
	g_list_foreach (tasklist->visible_tasks,
	                do_talika_task_update_visible_state, NULL);
}

void 
talika_tasklist_update_arrows (TalikaTaskList * tasklist)
{
	g_list_foreach (tasklist->visible_tasks,
	                do_talika_task_update_visible_state, NULL);
}

void
talika_tasklist_remove_launcher (TalikaTask *task)
{
	TalikaTaskList *tasklist = task->tasklist;

	tasklist->launchers = g_list_remove (tasklist->launchers, task);

	talika_launcher_free (task);
	
	talika_tasklist_save_launchers (tasklist);

	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

void
talika_tasklist_add_launcher (TalikaTask *task)
{
	gint sort_order = 0;
	WnckWindow *window = NULL;
	
	switch (task->type)
	{
		case TALIKA_TASK_WINDOW:
		{
			window = task->window;
		}
		break;

		case TALIKA_TASK_CLASS_GROUP:
		{
			
			window = ((TalikaTask *)(task->windows->data))->window;
		}
		break;

		default:
			break;
	}

	//get sort order
	sort_order = talika_task_get_sort_order (task);
	
	//get data from cache
	TalikaClass *class_obj = 
		talika_class_cache_get_class (task->tasklist->class_cache, window);

	if(class_obj->desktop_id)
	{
		gchar *spec = 
			g_strdup_printf ("%s:%s",class_obj->desktop_id,
			                 class_obj->class_name);
		
		TalikaTaskList *tasklist = task->tasklist;

		TalikaTask *launcher;
		launcher = talika_launcher_new (tasklist,spec);

		if(launcher)
		{
			talika_task_set_sort_order (launcher,sort_order);
			tasklist->launchers = g_list_append (tasklist->launchers,launcher);
		}

		talika_tasklist_save_launchers (tasklist);
		
		talika_tasklist_update_lists (tasklist);
		//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));

		g_free (spec);
	}
}

void 
talika_tasklist_update_launchers_list (TalikaTaskList *tasklist,
                                            GSList *new_list)
{
	//free old list
	if(tasklist->launchers_ids)
	{
		g_slist_foreach(tasklist->launchers_ids,(GFunc)g_free,NULL);
		g_slist_free (tasklist->launchers_ids);
	}

	tasklist->launchers_ids = new_list;

	//************** creating launchers  ***************//
	talika_tasklist_free_launchers (tasklist);

	gint n_elements = 0;
	GSList *lun;
	lun = tasklist->launchers_ids;

	while(lun)
	{
		gchar *spec = lun->data;

		TalikaTask *launcher;
		launcher = talika_launcher_new (tasklist,spec);

		if(launcher)
		{
			tasklist->launchers = g_list_append (tasklist->launchers,launcher);
			n_elements++;
		}

		lun = lun->next;
	}

	//assign order to launchers
	GList *l = tasklist->launchers;
	while (l)
	{
		TalikaTask *launcher = l->data;
		
		talika_task_set_sort_order (launcher, -n_elements);
		n_elements--;
		
		l = l->next;
	}
	
	//******************************************************//
	
	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
}

void
talika_tasklist_save_launchers (TalikaTaskList * tasklist)
{
	if(tasklist->launchers_ids)
	{
		g_slist_foreach(tasklist->launchers_ids,(GFunc)g_free,NULL);
		g_slist_free (tasklist->launchers_ids);
	}
	tasklist->launchers_ids = NULL;
	
	GList *l = g_list_copy (tasklist->launchers);
	l = g_list_sort (l, talika_tasklist_tasks_compare);
	
	while (l)
	{
		TalikaTask *launcher = l->data;

		gchar *str;

		if (launcher->icon_path)
			str = g_strdup_printf ("%s:%s:%s",launcher->desktop_id,
			                       launcher->class_name,launcher->icon_path);
		else
			str = g_strdup_printf ("%s:%s",launcher->desktop_id,
			                       launcher->class_name);
		
		tasklist->launchers_ids = g_slist_append(tasklist->launchers_ids,
		                                         str);
		
		l = l->next;
	}

	g_list_free (l);

	//save to gconf
	if (tasklist->applet)
	{
		panel_applet_gconf_set_list (PANEL_APPLET (tasklist->applet),
		                             "launchers",
		                             GCONF_VALUE_STRING,
		                             tasklist->launchers_ids,
		                             NULL);
	}
}

gboolean 
talika_tasklist_change_orient (TalikaTaskList * tasklist,
                               PanelAppletOrient orient)
{
	if (tasklist->panel_orient == orient)
		return FALSE;

	//check if recreation needed
	if (orient == PANEL_APPLET_ORIENT_DOWN &&
	    tasklist->panel_orient == PANEL_APPLET_ORIENT_UP)
	{
		tasklist->panel_orient = orient;
		return FALSE;
	}

	if (orient == PANEL_APPLET_ORIENT_LEFT &&
	    tasklist->panel_orient == PANEL_APPLET_ORIENT_RIGHT)
	{
		tasklist->panel_orient = orient;
		return FALSE;
	}

	if (orient == PANEL_APPLET_ORIENT_UP &&
	    tasklist->panel_orient == PANEL_APPLET_ORIENT_DOWN)
	{
		tasklist->panel_orient = orient;
		return FALSE;
	}

	if (orient == PANEL_APPLET_ORIENT_RIGHT &&
	    tasklist->panel_orient == PANEL_APPLET_ORIENT_LEFT)
	{
		tasklist->panel_orient = orient;
		return FALSE;
	}

	tasklist->panel_orient = orient;

	gtk_widget_destroy (tasklist->box);

	tasklist->box = get_talika_tasklist_box (orient);
	gtk_box_set_spacing(GTK_BOX(tasklist->box),tasklist->button_spacing);

	gtk_container_set_reallocate_redraws (GTK_CONTAINER
	                                      (tasklist->box), TRUE);
	gtk_widget_show_all (tasklist->box);

	//this is the hack to get right panel size
	g_signal_connect_after (tasklist->box, "expose_event",
	                        G_CALLBACK (talika_tasklist_expose),
	                        tasklist);

	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));

	return TRUE;
}

/*############################### Tasklist ###################################*/

//calculate the size tasklist requires from panel
void 
talika_tasklist_size_allocation (TalikaTaskList * tasklist, 
                                 int *size_hints)
{
	GList *l;
	GtkAllocation allocation;

	l = tasklist->visible_tasks;

	size_hints[0] = 0;
	//size_hints[1] = tasklist->panel_size;

	while (l != NULL)
	{
		TalikaTask *task = (TalikaTask *) l->data;
		l = l->next;

		allocation = task->button->allocation;

		if (tasklist->panel_orient == PANEL_APPLET_ORIENT_UP ||
		    tasklist->panel_orient == PANEL_APPLET_ORIENT_DOWN)
		{
			size_hints[0] += allocation.width + tasklist->button_spacing + 2;
		}
		else
		{
			size_hints[0] += allocation.height + tasklist->button_spacing + 2;
		}
	}

	//reserve space at least of panel size
	if (size_hints[0] == 0)
		size_hints[0] = tasklist->panel_size;
	else
	{
		//remove extra pixels from end
		size_hints[0] -= tasklist->button_spacing;
	}

	size_hints[1] = size_hints[0];

	//create background pixmap here
	GtkWidget *widget = tasklist->box;
	GdkGC *gc;

	if (GTK_WIDGET_DRAWABLE (widget))
	{
		/* get a screenshot of the background */

		if (tasklist->background != NULL)
			g_object_unref (tasklist->background);

		tasklist->background = gdk_pixmap_new (widget->window,
		                                       widget->allocation.
		                                       width,
		                                       widget->allocation.
		                                       height, -1);
		gc = gdk_gc_new (tasklist->background);
		gdk_draw_drawable (tasklist->background, gc,
		                   widget->window,
		                   widget->allocation.x,
		                   widget->allocation.y, 0, 0,
		                   widget->allocation.width,
		                   widget->allocation.height);

		g_object_unref (gc);
	}

	//update windows minimiation destinations
	talika_tasklist_update_icon_geometries (tasklist,
	                                        tasklist->visible_tasks);
}

//detect panel size change
static gboolean 
talika_tasklist_expose (GtkWidget * widget,
                        GdkEventExpose * event, gpointer data)
{
	TalikaTaskList *tasklist = data;
	guint panel_size;

	if (tasklist->panel_orient == PANEL_APPLET_ORIENT_UP ||
	    tasklist->panel_orient == PANEL_APPLET_ORIENT_DOWN)
	{
		panel_size = widget->allocation.height;
	}
	else
	{
		panel_size = widget->allocation.width;
	}

	if (tasklist->panel_size != panel_size)
	{
		tasklist->panel_size = panel_size;

		//reset icon cache
		talika_icon_cache_free(tasklist->icon_cache);
		tasklist->icon_cache = talika_icon_cache_new ();
		
		//update all launchers
		g_list_foreach (tasklist->launchers,
		                do_talika_launcher_update_visible_state, NULL);
		
		//update all buttons              
		g_list_foreach (tasklist->visible_tasks,
		                do_talika_task_update_visible_state, NULL);
	}

	return FALSE;
}

/* free tasklist */
void 
talika_tasklist_free (TalikaTaskList * tasklist)
{
	talika_tasklist_disconnect_screen (tasklist);
	talika_tasklist_free_tasks (tasklist);
	talika_tasklist_free_launchers (tasklist);
	
	g_hash_table_unref (tasklist->win_hash);
	g_hash_table_unref (tasklist->class_group_hash);
	gtk_widget_destroy (tasklist->box);

	if (tasklist->background)
	{
		g_object_unref (tasklist->background);
	}

	if (tasklist->activate_timeout_id != 0)
	{
		g_source_remove (tasklist->activate_timeout_id);
		tasklist->activate_timeout_id = 0;
	}

	if (tasklist->idle_callback_tag != 0)
	{
		g_source_remove (tasklist->idle_callback_tag);
		tasklist->idle_callback_tag = 0;
	}

	if (tasklist->enter_timeout_id != 0)
	{
		g_source_remove (tasklist->enter_timeout_id);
		tasklist->enter_timeout_id = 0;
	}

	if (tasklist->leave_timeout_id != 0)
	{
		g_source_remove (tasklist->leave_timeout_id);
		tasklist->leave_timeout_id = 0;
	}

	if (tasklist->popup)
	{
		talika_tooltips_free_popup (tasklist);
	}

	talika_preview_image_cache_free(tasklist->preview_cache);
	talika_class_cache_free(tasklist->class_cache);
	talika_icon_cache_free(tasklist->icon_cache);

	if(tasklist->launchers_ids)
	{
		g_slist_foreach(tasklist->launchers_ids,(GFunc)g_free,NULL);
		g_slist_free (tasklist->launchers_ids);
	}

	g_free (tasklist);
}

static GtkWidget* 
get_talika_tasklist_box (PanelAppletOrient orient)
{
	if (orient == PANEL_APPLET_ORIENT_UP ||
	    orient == PANEL_APPLET_ORIENT_DOWN)
	{
		return gtk_hbox_new (FALSE, 0);
	}
	else
	{
		return gtk_vbox_new (FALSE, 0);
	}
}

/* Create new tasklist for screen */
TalikaTaskList* 
talika_tasklist_new (WnckScreen * screen, guint size,
                     PanelAppletOrient orient)
{

	//change current directory to home
	g_chdir (g_get_home_dir ());

	//initilise stock icons
	talika_stock_icons_init();
	
	//modify button style by adding custom rc style 
	//this is to prevent image clipping on button
	gchar *style =  "style \"custom\"\n"
			"{\n"
			"xthickness = 1\n"
			"ythickness = 1\n"
			"GtkButton      ::child-displacement-x = 1\n"
			"GtkButton      ::child-displacement-y = 1\n"
			"GtkButton      ::default-border       = { 0, 0, 0, 0 }\n"
			"GtkWidget::focus_padding = 0\n"
			"}\n"
			"widget        \"*.tasklist-button\" 			style \"custom\"";

	gtk_rc_parse_string (style);


	//create new tasklist
	TalikaTaskList *tasklist = g_new0 (TalikaTaskList, 1);

	tasklist->screen = screen;
	tasklist->panel_size = size;	
	tasklist->panel_orient = orient;

	tasklist->box = get_talika_tasklist_box (orient);

	gtk_container_set_reallocate_redraws (GTK_CONTAINER
	                                      (tasklist->box), TRUE);
	gtk_widget_show_all (tasklist->box);

	tasklist->win_hash = g_hash_table_new (NULL, NULL);
	tasklist->class_group_hash = g_hash_table_new (g_str_hash ,g_str_equal);
	tasklist->class_cache = talika_class_cache_new ();
	tasklist->icon_cache = talika_icon_cache_new ();
	
	//intilise properties (used by test program, in applet 
	// gconf values are loaded later)
	tasklist->include_all_workspaces = FALSE;
	tasklist->include_all_monitors = TRUE;
	
	tasklist->use_grouping = TRUE;
	tasklist->group_reverse_order = FALSE;
	tasklist->group_middle_close = FALSE;
	
	tasklist->button_relief = FALSE;
	tasklist->toggle_current_task = TRUE;
	tasklist->dimm_minimised = FALSE;
	tasklist->mono_icons = FALSE;
	tasklist->show_theme_icons = TRUE;
	tasklist->running_indicator = TRUE;
	tasklist->icon_size_offset = 4;
	tasklist->button_spacing = 2;
	
	tasklist->enable_previews = TRUE;
	tasklist->previews_group_only = FALSE;
	tasklist->previews_group_click = FALSE;
	tasklist->previews_hide_activation = FALSE;
	tasklist->preview_rows = 2;
	tasklist->preview_cols = 4;
	tasklist->preview_size = 120;
	tasklist->preview_transparency = 7;
	tasklist->preview_timeout = 250;
	
	tasklist->preview_rounded_corners = TRUE;
	tasklist->preview_reverse_order = FALSE;

	gdk_color_parse ("#8a8a56563131",&tasklist->running_indicator_color);
	gdk_color_parse ("#000000000000",&tasklist->preview_popup_bg_color);
	gdk_color_parse ("#ffffffffffff",&tasklist->preview_popup_fg_color);
	gdk_color_parse ("#8a8a56563131",&tasklist->preview_win_border_color);
	gdk_color_parse ("#ffffffffffff",&tasklist->preview_win_title_color);
	
	//enabling it is not working, so keep it false
	tasklist->switch_workspace_on_unminimize = FALSE;

	tasklist->monitor_num = -1;
	tasklist->monitor_geometry.width = -1;

	tasklist->cursor_over_icon = FALSE;
	tasklist->hide_delay_started = FALSE;
	tasklist->popup = NULL;
	tasklist->enter_timeout_id = 0;
	tasklist->leave_timeout_id = 0;

	gtk_box_set_spacing(GTK_BOX(tasklist->box),tasklist->button_spacing);
	if (tasklist->enable_previews)
		tasklist->preview_cache = talika_preview_image_cache_new ();

	//add some launchers
	/*gchar *str;
	tasklist->launchers_ids = NULL;
	str = g_strdup("nautilus:File manager");
	tasklist->launchers_ids = g_slist_append(tasklist->launchers_ids,
	                                         str);
	str = g_strdup("firefox:Web browser");
	tasklist->launchers_ids = g_slist_append(tasklist->launchers_ids,
	                                         str);
	str = g_strdup("thunderbird-3.0:sherrder");
	tasklist->launchers_ids = g_slist_append(tasklist->launchers_ids,
	                                         str);*/

	//this is the hack to get right panel size
	g_signal_connect_after (tasklist->box, "expose_event",
	                        G_CALLBACK (talika_tasklist_expose),
	                        tasklist);

	//populate tasklist here
	talika_tasklist_update_lists (tasklist);
	talika_tasklist_connect_screen (tasklist);

	return tasklist;
}
